// A map between namespaces and base URLs for their online documentation
baseURLs = [
    [ 'GLib', 'https://docs.gtk.org/glib/' ],
    [ 'GObject', 'https://docs.gtk.org/gobject/' ],
    [ 'Gio', 'https://docs.gtk.org/gio/' ],
    [ 'GSSDP', 'https://gnome.pages.gitlab.gnome.org/gssdp/docs/' ],
    [ 'libxml2', 'https://gnome.pages.gitlab.gnome.org/libxml2/devhelp/libxml2-tree.html#']
]
